@@ -68,14 +68,14 @@ php_http_message_t *php_http_message_init_env(php_http_message_t *message, php_h
 				message->http.info.request.method = estrdup(Z_STRVAL_P(sval));
 			}
 			if ((sval = php_http_env_get_server_var(ZEND_STRL("REQUEST_URI"), 1))) {
-				message->http.info.request.url = php_http_url_parse(Z_STRVAL_P(sval), Z_STRLEN_P(sval), ~0);
+				message->http.info.request.url = php_http_url_parse(Z_STRVAL_P(sval), Z_STRLEN_P(sval), PHP_HTTP_URL_STDFLAGS);
 			}
 
 			php_http_env_get_request_headers(&message->hdrs);
 			break;
 
 		case PHP_HTTP_RESPONSE:
-			message = php_http_message_init(NULL, type, NULL);
+			message = php_http_message_init(message, type, NULL);
 			if (!SG(sapi_headers).http_status_line || !php_http_info_parse((php_http_info_t *) &message->http, SG(sapi_headers).http_status_line)) {
 				if (!(message->http.info.response.code = SG(sapi_headers).http_response_code)) {
 					message->http.info.response.code = 200;
@@ -177,24 +177,24 @@ zend_bool php_http_message_is_multipart(php_http_message_t *msg, char **boundary
 		popts.input.str = ct->val;
 		popts.input.len = ct->len;
 
-		if (php_http_params_parse(&params, &popts)) {
+		if (EXPECTED(php_http_params_parse(&params, &popts))) {
 			zval *cur, *arg;
 			zend_string *ct_str;
 			zend_ulong index;
 
 			zend_hash_internal_pointer_reset(&params);
 
-			if ((cur = zend_hash_get_current_data(&params))
+			if (EXPECTED((cur = zend_hash_get_current_data(&params))
 			&&	(Z_TYPE_P(cur) == IS_ARRAY)
-			&&	(HASH_KEY_IS_STRING == zend_hash_get_current_key(&params, &ct_str, &index))
+			&&	(HASH_KEY_IS_STRING == zend_hash_get_current_key(&params, &ct_str, &index)))
 			) {
 				if (php_http_match(ct_str->val, "multipart", PHP_HTTP_MATCH_WORD)) {
 					is_multipart = 1;
 
 					/* get boundary */
-					if (boundary
+					if (EXPECTED(boundary
 					&&	(arg = zend_hash_str_find(Z_ARRVAL_P(cur), ZEND_STRL("arguments")))
-					&&	Z_TYPE_P(arg) == IS_ARRAY
+					&&	Z_TYPE_P(arg) == IS_ARRAY)
 					) {
 						zval *val;
 						php_http_arrkey_t key;
@@ -204,7 +204,7 @@ zend_bool php_http_message_is_multipart(php_http_message_t *msg, char **boundary
 							if (key.key && key.key->len == lenof("boundary") && !strcasecmp(key.key->val, "boundary")) {
 								zend_string *bnd = zval_get_string(val);
 
-								if (bnd->len) {
+								if (EXPECTED(bnd->len)) {
 									*boundary = estrndup(bnd->val, bnd->len);
 								}
 								zend_string_release(bnd);
@@ -330,24 +330,14 @@ void php_http_message_update_headers(php_http_message_t *msg)
 static void message_headers(php_http_message_t *msg, php_http_buffer_t *str)
 {
 	char *tmp = NULL;
+	size_t len = 0;
 
-	switch (msg->type) {
-		case PHP_HTTP_REQUEST:
-			php_http_buffer_appendf(str, PHP_HTTP_INFO_REQUEST_FMT_ARGS(&msg->http, tmp, PHP_HTTP_CRLF));
-			PTR_FREE(tmp);
-			break;
-
-		case PHP_HTTP_RESPONSE:
-			php_http_buffer_appendf(str, PHP_HTTP_INFO_RESPONSE_FMT_ARGS(&msg->http, tmp, PHP_HTTP_CRLF));
-			PTR_FREE(tmp);
-			break;
-
-		default:
-			break;
-	}
-
+	php_http_info_to_string((php_http_info_t *) msg, &tmp, &len, PHP_HTTP_CRLF);
 	php_http_message_update_headers(msg);
+
+	php_http_buffer_append(str, tmp, len);
 	php_http_header_to_string(str, &msg->hdrs);
+	PTR_FREE(tmp);
 }
 
 void php_http_message_to_callback(php_http_message_t *msg, php_http_pass_callback_t cb, void *cb_arg)
@@ -415,7 +405,7 @@ php_http_message_t *php_http_message_reverse(php_http_message_t *msg)
 	if (c > 1) {
 		php_http_message_t *tmp = msg, **arr;
 
-		arr = ecalloc(c, sizeof(**arr));
+		arr = ecalloc(c, sizeof(*arr));
 		for (i = 0; i < c; ++i) {
 			arr[i] = tmp;
 			tmp = tmp->parent;
@@ -461,7 +451,7 @@ php_http_message_t *php_http_message_copy_ex(php_http_message_t *from, php_http_
 
 		copy = temp = php_http_message_init(to, 0, php_http_message_body_copy(from->body, NULL));
 		php_http_message_set_info(temp, &info);
-		zend_hash_copy(&temp->hdrs, &from->hdrs, (copy_ctor_func_t) zval_add_ref);
+		array_copy(&from->hdrs, &temp->hdrs);
 
 		if (parents) while (from->parent) {
 			info.type = from->parent->type;
@@ -481,7 +471,7 @@ php_http_message_t *php_http_message_copy_ex(php_http_message_t *from, php_http_
 
 void php_http_message_dtor(php_http_message_t *message)
 {
-	if (message) {
+	if (EXPECTED(message)) {
 		zend_hash_destroy(&message->hdrs);
 		php_http_message_body_free(&message->body);
 
@@ -503,7 +493,7 @@ void php_http_message_dtor(php_http_message_t *message)
 
 void php_http_message_free(php_http_message_t **message)
 {
-	if (*message) {
+	if (EXPECTED(*message)) {
 		if ((*message)->parent) {
 			php_http_message_free(&(*message)->parent);
 		}
@@ -519,8 +509,8 @@ zend_class_entry *php_http_message_get_class_entry(void)
 	return php_http_message_class_entry;
 }
 
-static zval *php_http_message_object_read_prop(zval *object, zval *member, int type, void **cache_slot, zval *rv);
-static void php_http_message_object_write_prop(zval *object, zval *member, zval *value, void **cache_slot);
+static zval *php_http_message_object_read_prop(zend_object *object, zend_string *member, int type, void **cache_slot, zval *rv);
+static zval *php_http_message_object_write_prop(zend_object *object, zend_string *member, zval *value, void **cache_slot);
 
 static zend_object_handlers php_http_message_object_handlers;
 static HashTable php_http_message_object_prophandlers;
@@ -548,12 +538,14 @@ static php_http_message_object_prophandler_t *php_http_message_object_get_propha
 	return zend_hash_str_find_ptr(&php_http_message_object_prophandlers, name_str->val, name_str->len);
 }
 static void php_http_message_object_prophandler_get_type(php_http_message_object_t *obj, zval *return_value) {
+	zval_ptr_dtor(return_value);
 	RETVAL_LONG(obj->message->type);
 }
 static void php_http_message_object_prophandler_set_type(php_http_message_object_t *obj, zval *value) {
 	php_http_message_set_type(obj->message, zval_get_long(value));
 }
 static void php_http_message_object_prophandler_get_request_method(php_http_message_object_t *obj, zval *return_value) {
+	zval_ptr_dtor(return_value);
 	if (PHP_HTTP_MESSAGE_TYPE(REQUEST, obj->message) && obj->message->http.info.request.method) {
 		RETVAL_STRING(obj->message->http.info.request.method);
 	} else {
@@ -571,6 +563,7 @@ static void php_http_message_object_prophandler_get_request_url(php_http_message
 	char *url_str;
 	size_t url_len;
 
+	zval_ptr_dtor(return_value);
 	if (PHP_HTTP_MESSAGE_TYPE(REQUEST, obj->message) && obj->message->http.info.request.url && php_http_url_to_string(obj->message->http.info.request.url, &url_str, &url_len, 0)) {
 		RETVAL_STR(php_http_cs2zs(url_str, url_len));
 	} else {
@@ -579,10 +572,11 @@ static void php_http_message_object_prophandler_get_request_url(php_http_message
 }
 static void php_http_message_object_prophandler_set_request_url(php_http_message_object_t *obj, zval *value) {
 	if (PHP_HTTP_MESSAGE_TYPE(REQUEST, obj->message)) {
-		PTR_SET(obj->message->http.info.request.url, php_http_url_from_zval(value, ~0));
+		PTR_SET(obj->message->http.info.request.url, php_http_url_from_zval(value, PHP_HTTP_URL_STDFLAGS));
 	}
 }
 static void php_http_message_object_prophandler_get_response_status(php_http_message_object_t *obj, zval *return_value) {
+	zval_ptr_dtor(return_value);
 	if (PHP_HTTP_MESSAGE_TYPE(RESPONSE, obj->message) && obj->message->http.info.response.status) {
 		RETVAL_STRING(obj->message->http.info.response.status);
 	} else {
@@ -597,6 +591,7 @@ static void php_http_message_object_prophandler_set_response_status(php_http_mes
 	}
 }
 static void php_http_message_object_prophandler_get_response_code(php_http_message_object_t *obj, zval *return_value) {
+	zval_ptr_dtor(return_value);
 	if (PHP_HTTP_MESSAGE_TYPE(RESPONSE, obj->message)) {
 		RETVAL_LONG(obj->message->http.info.response.code);
 	} else {
@@ -613,6 +608,7 @@ static void php_http_message_object_prophandler_get_http_version(php_http_messag
 	char *version_str;
 	size_t version_len;
 
+	zval_ptr_dtor(return_value);
 	php_http_version_to_string(&obj->message->http.version, &version_str, &version_len, NULL, NULL);
 	RETVAL_STR(php_http_cs2zs(version_str, version_len));
 }
@@ -622,38 +618,55 @@ static void php_http_message_object_prophandler_set_http_version(php_http_messag
 	zend_string_release(zs);
 }
 static void php_http_message_object_prophandler_get_headers(php_http_message_object_t *obj, zval *return_value ) {
+	zval tmp;
+
+	ZVAL_COPY_VALUE(&tmp, return_value);
 	array_init(return_value);
 	array_copy(&obj->message->hdrs, Z_ARRVAL_P(return_value));
+	zval_ptr_dtor(&tmp);
 }
 static void php_http_message_object_prophandler_set_headers(php_http_message_object_t *obj, zval *value) {
-	HashTable *headers;
-	zval *orig_value = value;
+	int converted = 0;
+	HashTable garbage, *src;
 
 	if (Z_TYPE_P(value) != IS_ARRAY && Z_TYPE_P(value) != IS_OBJECT) {
-		convert_to_array_ex(value);
+		converted = 1;
+		SEPARATE_ZVAL(value);
+		convert_to_array(value);
 	}
-	headers = HASH_OF(value);
+	src = HASH_OF(value);
 
-	zend_hash_clean(&obj->message->hdrs);
-	array_copy(headers, &obj->message->hdrs);
+	garbage = obj->message->hdrs;
+	zend_hash_init(&obj->message->hdrs, zend_hash_num_elements(src), NULL, ZVAL_PTR_DTOR, 0);
+	array_copy(HASH_OF(value), &obj->message->hdrs);
 
-	if (orig_value != value) {
+	zend_hash_destroy(&garbage);
+
+	if (converted) {
 		zval_ptr_dtor(value);
 	}
 }
 static void php_http_message_object_prophandler_get_body(php_http_message_object_t *obj, zval *return_value) {
-	if (obj->body) {
-		RETVAL_OBJECT(&obj->body->zo, 1);
-	} else {
-		RETVAL_NULL();
+	zval tmp;
+
+	if (!obj->body) {
+		RETURN_NULL();
 	}
+
+	ZVAL_COPY_VALUE(&tmp, return_value);
+	RETVAL_OBJECT(&obj->body->zo, 1);
+	zval_ptr_dtor(&tmp);
 }
 static void php_http_message_object_prophandler_set_body(php_http_message_object_t *obj, zval *value) {
 	php_http_message_object_set_body(obj, value);
 }
 static void php_http_message_object_prophandler_get_parent_message(php_http_message_object_t *obj, zval *return_value) {
 	if (obj->message->parent) {
+		zval tmp;
+
+		ZVAL_COPY_VALUE(&tmp, return_value);
 		RETVAL_OBJECT(&obj->parent->zo, 1);
+		zval_ptr_dtor(&tmp);
 	} else {
 		RETVAL_NULL();
 	}
@@ -662,10 +675,10 @@ static void php_http_message_object_prophandler_set_parent_message(php_http_mess
 	if (Z_TYPE_P(value) == IS_OBJECT && instanceof_function(Z_OBJCE_P(value), php_http_message_class_entry)) {
 		php_http_message_object_t *parent_obj = PHP_HTTP_OBJ(NULL, value);
 
-		if (obj->message->parent) {
-			zend_objects_store_del(&obj->parent->zo);
-		}
 		Z_ADDREF_P(value);
+		if (obj->message->parent) {
+			zend_object_release(&obj->parent->zo);
+		}
 		obj->parent = parent_obj;
 		obj->message->parent = parent_obj->message;
 	}
@@ -675,6 +688,8 @@ static void php_http_message_object_prophandler_set_parent_message(php_http_mess
 	do { \
 		if (!obj->message) { \
 			obj->message = php_http_message_init(NULL, 0, NULL); \
+		} else if (!obj->body && php_http_message_body_size(obj->message->body)) { \
+			php_http_message_object_init_body_object(obj); \
 		} \
 	} while(0)
 
@@ -693,7 +708,7 @@ void php_http_message_object_reverse(zval *zmsg, zval *return_value)
 		php_http_message_object_t **objects;
 		int last;
 
-		objects = ecalloc(i, sizeof(**objects));
+		objects = ecalloc(i, sizeof(*objects));
 
 		/* we are the first message */
 		objects[0] = obj;
@@ -763,7 +778,7 @@ ZEND_RESULT_CODE php_http_message_object_set_body(php_http_message_object_t *msg
 		case IS_RESOURCE:
 			php_stream_from_zval_no_verify(s, zbody);
 			if (!s) {
-				php_http_throw(unexpected_val, "The stream is not a valid resource", NULL);
+				php_http_throw(unexpected_val, "The stream is not a valid resource");
 				return FAILURE;
 			}
 
@@ -797,7 +812,7 @@ ZEND_RESULT_CODE php_http_message_object_set_body(php_http_message_object_t *msg
 		body_obj->body = php_http_message_body_init(NULL, NULL);
 	}
 	if (msg_obj->body) {
-		zend_objects_store_del(&msg_obj->body->zo);
+		zend_object_release(&msg_obj->body->zo);
 	}
 	if (msg_obj->message) {
 		php_http_message_body_free(&msg_obj->message->body);
@@ -843,10 +858,10 @@ php_http_message_object_t *php_http_message_object_new_ex(zend_class_entry *ce,
 	return o;
 }
 
-zend_object *php_http_message_object_clone(zval *this_ptr)
+zend_object *php_http_message_object_clone(zend_object *this_ptr)
 {
-	php_http_message_object_t *new_obj = NULL;
-	php_http_message_object_t *old_obj = PHP_HTTP_OBJ(NULL, this_ptr);
+	php_http_message_object_t *new_obj;
+	php_http_message_object_t *old_obj = PHP_HTTP_OBJ(this_ptr, NULL);
 
 	new_obj = php_http_message_object_new_ex(old_obj->zo.ce, php_http_message_copy(old_obj->message, NULL));
 	zend_objects_clone_members(&new_obj->zo, &old_obj->zo);
@@ -858,6 +873,8 @@ void php_http_message_object_free(zend_object *object)
 {
 	php_http_message_object_t *o = PHP_HTTP_OBJ(object, NULL);
 
+	PTR_FREE(o->gc);
+
 	if (!Z_ISUNDEF(o->iterator)) {
 		zval_ptr_dtor(&o->iterator);
 		ZVAL_UNDEF(&o->iterator);
@@ -869,82 +886,63 @@ void php_http_message_object_free(zend_object *object)
 		o->message = NULL;
 	}
 	if (o->parent) {
-		if (GC_REFCOUNT(&o->parent->zo) == 1) {
-			zend_objects_store_del(&o->parent->zo);
-		}
-		zend_objects_store_del(&o->parent->zo);
+		zend_object_release(&o->parent->zo);
 		o->parent = NULL;
 	}
 	if (o->body) {
-		if (GC_REFCOUNT(&o->body->zo) == 1) {
-			zend_objects_store_del(&o->body->zo);
-		}
-		zend_objects_store_del(&o->body->zo);
+		zend_object_release(&o->body->zo);
 		o->body = NULL;
 	}
 	zend_object_std_dtor(object);
 }
 
-static zval *php_http_message_object_read_prop(zval *object, zval *member, int type, void **cache_slot, zval *tmp)
+static zval *php_http_message_object_get_prop_ptr(zend_object *object, zend_string *member, int type, void **cache_slot)
 {
-	zval *return_value;
-	zend_string *member_name = zval_get_string(member);
-	php_http_message_object_prophandler_t *handler = php_http_message_object_get_prophandler(member_name);
-
-	if (!handler || type == BP_VAR_R || type == BP_VAR_IS) {
-		return_value = zend_get_std_object_handlers()->read_property(object, member, type, cache_slot, tmp);
-
-		if (handler) {
-			php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, object);
-
-			PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
-			handler->read(obj, tmp);
-
-			zval_ptr_dtor(return_value);
-			ZVAL_COPY_VALUE(return_value, tmp);
-		}
-		zend_string_release(member_name);
-		return return_value;
-	} else {
-		php_property_proxy_t *proxy;
-		php_property_proxy_object_t *proxy_obj;
-
-		proxy = php_property_proxy_init(object, member_name);
-		proxy_obj = php_property_proxy_object_new_ex(NULL, proxy);
-
-		ZVAL_OBJ(tmp, &proxy_obj->zo);
-		zend_string_release(member_name);
-		return tmp;
-	}
+	return NULL;
 }
 
-static void php_http_message_object_write_prop(zval *object, zval *member, zval *value, void **cache_slot)
+static zval *php_http_message_object_read_prop(zend_object *object, zend_string *member, int type, void **cache_slot, zval *tmp)
 {
-	php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, object);
+	zval *return_value;
+	php_http_message_object_prophandler_t *handler = php_http_message_object_get_prophandler(member);
+
+	return_value = zend_get_std_object_handlers()->read_property(object, member, type, cache_slot, tmp);
+
+	if (handler && handler->read) {
+		php_http_message_object_t *obj = PHP_HTTP_OBJ(object, NULL);
+
+		handler->read(obj, return_value);
+	}
+	return return_value;
+}
+
+static zval *php_http_message_object_write_prop(zend_object *object, zend_string *member, zval *value, void **cache_slot)
+{
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(object, NULL);
 	php_http_message_object_prophandler_t *handler;
-	zend_string *member_name = zval_get_string(member);
 
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
-	if ((handler = php_http_message_object_get_prophandler(member_name))) {
+	if ((handler = php_http_message_object_get_prophandler(member))) {
 		handler->write(obj, value);
 	} else {
 		zend_get_std_object_handlers()->write_property(object, member, value, cache_slot);
 	}
-
-	zend_string_release(member_name);
+	return value;
 }
 
-static HashTable *php_http_message_object_get_debug_info(zval *object, int *is_temp)
+static HashTable *php_http_message_object_get_debug_info(zend_object *object, int *is_temp)
 {
-	zval tmp;
-	php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, object);
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(object, NULL);
 	HashTable *props = zend_get_std_object_handlers()->get_properties(object);
 	char *ver_str, *url_str = NULL;
 	size_t ver_len, url_len = 0;
+	zval tmp;
 
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
-	*is_temp = 0;
+	if (is_temp) {
+		*is_temp = 0;
+	}
 
 #define UPDATE_PROP(name_str, action_with_tmp) \
 	do { \
@@ -1014,6 +1012,56 @@ static HashTable *php_http_message_object_get_debug_info(zval *object, int *is_t
 	return props;
 }
 
+static HashTable *php_http_message_object_get_gc(zend_object *object, zval **table, int *n)
+{
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(object, NULL);
+	HashTable *props = object->handlers->get_properties(object);
+	uint32_t count = 2 + zend_hash_num_elements(props);
+	zval *val;
+
+	*n = 0;
+	*table = obj->gc = erealloc(obj->gc, count * sizeof(zval));
+
+	if (obj->body) {
+		ZVAL_OBJ(&obj->gc[(*n)++], &obj->body->zo);
+	}
+	if (obj->parent) {
+		ZVAL_OBJ(&obj->gc[(*n)++], &obj->parent->zo);
+	}
+
+	ZEND_HASH_FOREACH_VAL(props, val)
+	{
+		ZVAL_COPY_VALUE(&obj->gc[(*n)++], val);
+	}
+	ZEND_HASH_FOREACH_END();
+
+	return NULL;
+}
+
+static int php_http_message_object_cast(zend_object *object, zval *return_value, int type)
+{
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(object, NULL);
+	char *string;
+	size_t length;
+
+	switch (type) {
+	case IS_STRING:
+		PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
+		php_http_message_to_string(obj->message, &string, &length);
+		if (string) {
+			RETVAL_STR(php_http_cs2zs(string, length));
+		} else {
+			RETVAL_EMPTY_STRING();
+		}
+		return SUCCESS;
+	case _IS_BOOL:
+		RETVAL_TRUE;
+		return SUCCESS;
+	default:
+		return FAILURE;
+	}
+}
+
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage___construct, 0, 0, 0)
 	ZEND_ARG_INFO(0, message)
 	ZEND_ARG_INFO(0, greedy)
@@ -1045,7 +1093,7 @@ static PHP_METHOD(HttpMessage, __construct)
 			php_http_buffer_init_ex(&buf, 0x1000, PHP_HTTP_BUFFER_INIT_PREALLOC);
 			if (PHP_HTTP_MESSAGE_PARSER_STATE_FAILURE == php_http_message_parser_parse_stream(&p, &buf, s, flags, &msg)) {
 				if (!EG(exception)) {
-					php_http_throw(bad_message, "Could not parse message from stream", NULL);
+					php_http_throw(bad_message, "Could not parse message from stream");
 				}
 			}
 			php_http_buffer_dtor(&buf);
@@ -1053,7 +1101,7 @@ static PHP_METHOD(HttpMessage, __construct)
 		}
 
 		if (!msg && !EG(exception)) {
-			php_http_throw(bad_message, "Empty message received from stream", NULL);
+			php_http_throw(bad_message, "Empty message received from stream");
 		}
 	} else if (zmessage) {
 		zend_string *zs_msg = zval_get_string(zmessage);
@@ -1061,7 +1109,7 @@ static PHP_METHOD(HttpMessage, __construct)
 		msg = php_http_message_parse(NULL, zs_msg->val, zs_msg->len, greedy);
 
 		if (!msg && !EG(exception)) {
-			php_http_throw(bad_message, "Could not parse message: %.*s", MIN(25, zs_msg->len), zs_msg->val);
+			php_http_throw(bad_message, "Could not parse message: %.*s", (int) MIN(25, zs_msg->len), zs_msg->val);
 		}
 		zend_string_release(zs_msg);
 	}
@@ -1090,7 +1138,6 @@ static PHP_METHOD(HttpMessage, getBody)
 
 	if (!obj->body) {
 		php_http_message_object_init_body_object(obj);
-
 	}
 	if (obj->body) {
 		RETVAL_OBJECT(&obj->body->zo, 1);
@@ -1234,6 +1281,38 @@ static PHP_METHOD(HttpMessage, setHeaders)
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
 
+static inline void php_http_message_object_add_header(php_http_message_object_t *obj, const char *name_str, size_t name_len, zval *zvalue)
+{
+	char *name = php_http_pretty_key(estrndup(name_str, name_len), name_len, 1, 1);
+	zend_string *hstr, *vstr;
+	zval *header, tmp;
+
+	if (Z_TYPE_P(zvalue) == IS_NULL) {
+		return;
+	}
+
+	vstr = php_http_header_value_to_string(zvalue);
+
+	if ((name_len != lenof("Set-Cookie") && strcmp(name, "Set-Cookie"))
+	&&	(hstr = php_http_message_header_string(obj->message, name, name_len))) {
+		char *hdr_str;
+		size_t hdr_len = spprintf(&hdr_str, 0, "%s, %s", hstr->val, vstr->val);
+
+		ZVAL_STR(&tmp, php_http_cs2zs(hdr_str, hdr_len));
+		zend_symtable_str_update(&obj->message->hdrs, name, name_len, &tmp);
+		zend_string_release(hstr);
+		zend_string_release(vstr);
+	} else if ((header = php_http_message_header(obj->message, name, name_len))) {
+		convert_to_array(header);
+		ZVAL_STR(&tmp, vstr);
+		zend_hash_next_index_insert(Z_ARRVAL_P(header), &tmp);
+	} else {
+		ZVAL_STR(&tmp, vstr);
+		zend_symtable_str_update(&obj->message->hdrs, name, name_len, &tmp);
+	}
+	efree(name);
+}
+
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_addHeader, 0, 0, 2)
 	ZEND_ARG_INFO(0, header)
 	ZEND_ARG_INFO(0, value)
@@ -1246,19 +1325,10 @@ static PHP_METHOD(HttpMessage, addHeader)
 
 	if (SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "sz", &name_str, &name_len, &zvalue)) {
 		php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
-		char *name = php_http_pretty_key(estrndup(name_str, name_len), name_len, 1, 1);
-		zval *header;
 
 		PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
-		Z_TRY_ADDREF_P(zvalue);
-		if ((header = php_http_message_header(obj->message, name, name_len))) {
-			convert_to_array(header);
-			zend_hash_next_index_insert(Z_ARRVAL_P(header), zvalue);
-		} else {
-			zend_symtable_str_update(&obj->message->hdrs, name, name_len, zvalue);
-		}
-		efree(name);
+		php_http_message_object_add_header(obj, name_str, name_len, zvalue);
 	}
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
@@ -1277,7 +1347,20 @@ static PHP_METHOD(HttpMessage, addHeaders)
 
 		PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
-		array_join(Z_ARRVAL_P(new_headers), &obj->message->hdrs, append, ARRAY_JOIN_STRONLY|ARRAY_JOIN_PRETTIFY);
+		if (append) {
+			php_http_arrkey_t key = {0};
+			zval *val;
+
+			ZEND_HASH_FOREACH_KEY_VAL(Z_ARRVAL_P(new_headers), key.h, key.key, val)
+			{
+				php_http_arrkey_stringify(&key, NULL);
+				php_http_message_object_add_header(obj, key.key->val, key.key->len, val);
+				php_http_arrkey_dtor(&key);
+			}
+			ZEND_HASH_FOREACH_END();
+		} else {
+			array_join(Z_ARRVAL_P(new_headers), &obj->message->hdrs, 0, ARRAY_JOIN_PRETTIFY|ARRAY_JOIN_STRONLY);
+		}
 	}
 	RETVAL_ZVAL(getThis(), 1, 0);
 }
@@ -1317,25 +1400,12 @@ ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, getInfo)
 {
 	if (SUCCESS == zend_parse_parameters_none()) {
-		char *str, *tmp = NULL;
-		size_t len;
+		char *str = NULL;
+		size_t len = 0;
 		php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
 
 		PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
-
-		switch (obj->message->type) {
-			case PHP_HTTP_REQUEST:
-				len = spprintf(&str, 0, PHP_HTTP_INFO_REQUEST_FMT_ARGS(&obj->message->http, tmp, ""));
-				PTR_FREE(tmp);
-				break;
-			case PHP_HTTP_RESPONSE:
-				len = spprintf(&str, 0, PHP_HTTP_INFO_RESPONSE_FMT_ARGS(&obj->message->http, tmp, ""));
-				PTR_FREE(tmp);
-				break;
-			default:
-				RETURN_NULL();
-				break;
-		}
+		php_http_info_to_string((php_http_info_t *) obj->message, &str, &len, "");
 
 		RETVAL_STR(php_http_cs2zs(str, len));
 	}
@@ -1415,7 +1485,7 @@ static PHP_METHOD(HttpMessage, getResponseCode)
 		PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 		if (obj->message->type != PHP_HTTP_RESPONSE) {
-			php_error_docref(NULL, E_WARNING, "http\\Message is not if type response");
+			php_error_docref(NULL, E_WARNING, "http\\Message is not of type response");
 			RETURN_FALSE;
 		}
 
@@ -1439,7 +1509,7 @@ static PHP_METHOD(HttpMessage, setResponseCode)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (obj->message->type != PHP_HTTP_RESPONSE) {
-		php_http_throw(bad_method_call, "http\\Message is not of type response", NULL);
+		php_http_throw(bad_method_call, "http\\Message is not of type response");
 		return;
 	}
 
@@ -1491,7 +1561,7 @@ static PHP_METHOD(HttpMessage, setResponseStatus)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (obj->message->type != PHP_HTTP_RESPONSE) {
-		php_http_throw(bad_method_call, "http\\Message is not of type response", NULL);
+		php_http_throw(bad_method_call, "http\\Message is not of type response");
 	}
 
 	PTR_SET(obj->message->http.info.response.status, estrndup(status, status_len));
@@ -1536,12 +1606,12 @@ static PHP_METHOD(HttpMessage, setRequestMethod)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (obj->message->type != PHP_HTTP_REQUEST) {
-		php_http_throw(bad_method_call, "http\\Message is not of type request", NULL);
+		php_http_throw(bad_method_call, "http\\Message is not of type request");
 		return;
 	}
 
 	if (method_len < 1) {
-		php_http_throw(invalid_arg, "Cannot set http\\Message's request method to an empty string", NULL);
+		php_http_throw(invalid_arg, "Cannot set http\\Message's request method to an empty string");
 		return;
 	}
 
@@ -1592,17 +1662,17 @@ static PHP_METHOD(HttpMessage, setRequestUrl)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (obj->message->type != PHP_HTTP_REQUEST) {
-		php_http_throw(bad_method_call, "http\\Message is not of type request", NULL);
+		php_http_throw(bad_method_call, "http\\Message is not of type request");
 		return;
 	}
 
 	zend_replace_error_handling(EH_THROW, php_http_get_exception_bad_url_class_entry(), &zeh);
-	url = php_http_url_from_zval(zurl, ~0);
+	url = php_http_url_from_zval(zurl, PHP_HTTP_URL_STDFLAGS);
 	zend_restore_error_handling(&zeh);
 
 	if (url && php_http_url_is_empty(url)) {
 		php_http_url_free(&url);
-		php_http_throw(invalid_arg, "Cannot set http\\Message's request url to an empty string", NULL);
+		php_http_throw(invalid_arg, "Cannot set http\\Message's request url to an empty string");
 	} else if (url) {
 		PTR_SET(obj->message->http.info.request.url, url);
 	}
@@ -1622,15 +1692,13 @@ static PHP_METHOD(HttpMessage, getParentMessage)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (!obj->message->parent) {
-		php_http_throw(unexpected_val, "http\\Message has not parent message", NULL);
+		php_http_throw(unexpected_val, "http\\Message has no parent message");
 		return;
 	}
 
 	RETVAL_OBJECT(&obj->parent->zo, 1);
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage___toString, 0, 0, 0)
-ZEND_END_ARG_INFO();
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_toString, 0, 0, 0)
 	ZEND_ARG_INFO(0, include_parent)
 ZEND_END_ARG_INFO();
@@ -1692,10 +1760,62 @@ static PHP_METHOD(HttpMessage, toCallback)
 		zend_fcall_info_args_clear(&fcd.fci, 1);
 		zval_ptr_dtor(&fcd.fcz);
 
-		RETURN_ZVAL(getThis(), 1, 0);
+		RETURN_ZVAL(&fcd.fcz, 1, 0);
 	}
 }
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessage___serialize, 0, 0, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+static PHP_METHOD(HttpMessage, __serialize)
+{
+	zend_ulong num_index;
+	zend_string *str_index;
+	zend_property_info *pi;
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
+	HashTable *props = php_http_message_object_get_debug_info(&obj->zo, NULL);
+
+	zend_parse_parameters_none();
+
+	array_init(return_value);
+
+	ZEND_HASH_FOREACH_KEY_PTR(&obj->zo.ce->properties_info, num_index, str_index, pi)
+	{
+		(void)num_index;
+		zval *val;
+		if (str_index && (val = zend_hash_find_ind(props, pi->name))) {
+			Z_TRY_ADDREF_P(val);
+			zend_hash_update(Z_ARRVAL_P(return_value), str_index, val);
+		}
+	}
+	ZEND_HASH_FOREACH_END();
+}
+
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessage___unserialize, 0, 1, IS_VOID, 0)
+	ZEND_ARG_TYPE_INFO(0, data, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+static PHP_METHOD(HttpMessage, __unserialize)
+{
+	HashTable *arr;
+	zend_string *key;
+	zval *val;
+	php_http_message_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
+
+	php_http_expect(SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "h", &arr), invalid_arg, return);
+
+	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
+
+	ZEND_HASH_FOREACH_STR_KEY_VAL(arr, key, val)
+	{
+		php_http_message_object_prophandler_t *ph = php_http_message_object_get_prophandler(key);
+		if (ph) {
+			ph->write(obj, val);
+		} else {
+			zend_update_property_ex(php_http_message_class_entry, &obj->zo, key, val);
+		}
+	}
+	ZEND_HASH_FOREACH_END();
+}
+
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_serialize, 0, 0, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, serialize)
@@ -1779,7 +1899,7 @@ static PHP_METHOD(HttpMessage, prepend)
 	for (msg[0] = obj->message; msg[0]; msg[0] = msg[0]->parent) {
 		for (msg[1] = prepend_obj->message; msg[1]; msg[1] = msg[1]->parent) {
 			if (msg[0] == msg[1]) {
-				php_http_throw(unexpected_val, "Cannot prepend a message located within the same message chain", NULL);
+				php_http_throw(unexpected_val, "Cannot prepend a message located within the same message chain");
 				return;
 			}
 		}
@@ -1839,7 +1959,7 @@ static PHP_METHOD(HttpMessage, splitMultipartBody)
 	PHP_HTTP_MESSAGE_OBJECT_INIT(obj);
 
 	if (!php_http_message_is_multipart(obj->message, &boundary)) {
-		php_http_throw(bad_method_call, "http\\Message is not a multipart message", NULL);
+		php_http_throw(bad_method_call, "http\\Message is not a multipart message");
 		return;
 	}
 
@@ -1850,7 +1970,7 @@ static PHP_METHOD(HttpMessage, splitMultipartBody)
 	RETURN_OBJ(&php_http_message_object_new_ex(obj->zo.ce, msg)->zo);
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_count, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpMessage_count, 0, 0, IS_LONG, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, count)
 {
@@ -1865,7 +1985,7 @@ static PHP_METHOD(HttpMessage, count)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_rewind, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpMessage_rewind, 0, 0, IS_VOID, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, rewind)
 {
@@ -1880,7 +2000,7 @@ static PHP_METHOD(HttpMessage, rewind)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_valid, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpMessage_valid, 0, 0, _IS_BOOL, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, valid)
 {
@@ -1891,7 +2011,7 @@ static PHP_METHOD(HttpMessage, valid)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_next, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpMessage_next, 0, 0, IS_VOID, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, next)
 {
@@ -1915,7 +2035,7 @@ static PHP_METHOD(HttpMessage, next)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_key, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_HttpMessage_key, 0, 0, IS_LONG, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, key)
 {
@@ -1926,7 +2046,7 @@ static PHP_METHOD(HttpMessage, key)
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessage_current, 0, 0, 0)
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_OBJ_INFO_EX(ai_HttpMessage_current, 0, 0, http\\Message, 0)
 ZEND_END_ARG_INFO();
 static PHP_METHOD(HttpMessage, current)
 {
@@ -1939,8 +2059,11 @@ static PHP_METHOD(HttpMessage, current)
 	}
 }
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessage___toString, 0, 0, IS_STRING, 0)
+ZEND_END_ARG_INFO();
+
 static zend_function_entry php_http_message_methods[] = {
-	PHP_ME(HttpMessage, __construct,        ai_HttpMessage___construct,        ZEND_ACC_PUBLIC|ZEND_ACC_CTOR)
+	PHP_ME(HttpMessage, __construct,        ai_HttpMessage___construct,        ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessage, getBody,            ai_HttpMessage_getBody,            ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessage, setBody,            ai_HttpMessage_setBody,            ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessage, addBody,            ai_HttpMessage_addBody,            ZEND_ACC_PUBLIC)
@@ -1975,6 +2098,8 @@ static zend_function_entry php_http_message_methods[] = {
 	/* implements Serializable */
 	PHP_ME(HttpMessage, serialize,          ai_HttpMessage_serialize,          ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessage, unserialize,        ai_HttpMessage_unserialize,        ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessage, __serialize,        ai_HttpMessage___serialize,        ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessage, __unserialize,      ai_HttpMessage___unserialize,      ZEND_ACC_PUBLIC)
 
 	/* implements Iterator */
 	PHP_ME(HttpMessage, rewind,             ai_HttpMessage_rewind,             ZEND_ACC_PUBLIC)
@@ -2009,9 +2134,11 @@ PHP_MINIT_FUNCTION(http_message)
 	php_http_message_object_handlers.read_property = php_http_message_object_read_prop;
 	php_http_message_object_handlers.write_property = php_http_message_object_write_prop;
 	php_http_message_object_handlers.get_debug_info = php_http_message_object_get_debug_info;
-	php_http_message_object_handlers.get_property_ptr_ptr = NULL;
+	php_http_message_object_handlers.get_property_ptr_ptr = php_http_message_object_get_prop_ptr;
+	php_http_message_object_handlers.get_gc = php_http_message_object_get_gc;
+	php_http_message_object_handlers.cast_object = php_http_message_object_cast;
 
-	zend_class_implements(php_http_message_class_entry, 3, spl_ce_Countable, zend_ce_serializable, zend_ce_iterator);
+	zend_class_implements(php_http_message_class_entry, 3, zend_ce_countable, zend_ce_serializable, zend_ce_iterator);
 
 	zend_hash_init(&php_http_message_object_prophandlers, 9, NULL, php_http_message_object_prophandler_hash_dtor, 1);
 	zend_declare_property_long(php_http_message_class_entry, ZEND_STRL("type"), PHP_HTTP_NONE, ZEND_ACC_PROTECTED);
