@@ -12,7 +12,7 @@
 
 #include "php_http_api.h"
 
-#include <ext/standard/php_lcg.h>
+#include "ext/standard/php_lcg.h"
 
 #define BOUNDARY_OPEN(body) \
 	do {\
@@ -40,18 +40,17 @@ php_http_message_body_t *php_http_message_body_init(php_http_message_body_t **bo
 		php_http_message_body_addref(body);
 		return body;
 	}
-	
+
 	body = ecalloc(1, sizeof(php_http_message_body_t));
 	body->refcount = 1;
 
 	if (stream) {
 		body->res = stream->res;
-		++GC_REFCOUNT(body->res);
+		GC_ADDREF(body->res);
 	} else {
-		stream = php_stream_temp_create(TEMP_STREAM_DEFAULT, 0xffff);
-		body->res = stream->res;
+		body->res = php_stream_temp_create(TEMP_STREAM_DEFAULT, 0xffff)->res;
 	}
-	php_stream_auto_cleanup(stream);
+	php_stream_auto_cleanup(php_http_message_body_stream(body));
 
 	if (body_ptr) {
 		*body_ptr = body;
@@ -91,9 +90,9 @@ void php_http_message_body_free(php_http_message_body_t **body_ptr)
 {
 	if (*body_ptr) {
 		php_http_message_body_t *body = *body_ptr;
-
 		if (!--body->refcount) {
 			zend_list_delete(body->res);
+			body->res = NULL;
 			PTR_FREE(body->boundary);
 			efree(body);
 		}
@@ -215,7 +214,7 @@ size_t php_http_message_body_append(php_http_message_body_t *body, const char *b
 	written = php_stream_write(s, buf, len);
 
 	if (written != len) {
-		php_error_docref(NULL, E_WARNING, "Failed to append %zu bytes to body; wrote %zu", len, written);
+		php_error_docref(NULL, E_WARNING, "Failed to append %zu bytes to body; wrote %zu", len, written == (size_t) -1 ? 0 : written);
 	}
 
 	return len;
@@ -263,9 +262,14 @@ void php_http_message_body_add_part(php_http_message_body_t *body, php_http_mess
 
 ZEND_RESULT_CODE php_http_message_body_add_form_field(php_http_message_body_t *body, const char *name, const char *value_str, size_t value_len)
 {
-	zend_string *safe_name = zend_string_init(name, strlen(name), 0);
+	zend_string *safe_name, *zstr_name = zend_string_init(name, strlen(name), 0);
 
-	safe_name = php_addslashes(safe_name, 1);
+#if PHP_VERSION_ID < 70300
+	safe_name = php_addslashes(zstr_name, 1);
+#else
+	safe_name = php_addslashes(zstr_name);
+	zend_string_release_ex(zstr_name, 0);
+#endif
 
 	BOUNDARY_OPEN(body);
 	php_http_message_body_appendf(
@@ -285,9 +289,14 @@ ZEND_RESULT_CODE php_http_message_body_add_form_file(php_http_message_body_t *bo
 {
 	size_t path_len = strlen(path);
 	char *path_dup = estrndup(path, path_len);
-	zend_string *base_name, *safe_name = zend_string_init(name, strlen(name), 0);
+	zend_string *base_name, *safe_name, *zstr_name = zend_string_init(name, strlen(name), 0);
 
-	safe_name = php_addslashes(safe_name, 1);
+#if PHP_VERSION_ID < 70300
+	safe_name = php_addslashes(zstr_name, 1);
+#else
+	safe_name = php_addslashes(zstr_name);
+	zend_string_release_ex(zstr_name, 0);
+#endif
 	base_name = php_basename(path_dup, path_len, NULL, 0);
 
 	BOUNDARY_OPEN(body);
@@ -321,7 +330,7 @@ static inline char *format_key(php_http_arrkey_t *key, const char *prefix) {
 	} else if (key->key) {
 		new_key = estrdup(key->key->val);
 	} else {
-		new_key = estrdup("");
+		spprintf(&new_key, 0, "%lu", key->h);
 	}
 
 	return new_key;
@@ -340,8 +349,8 @@ static ZEND_RESULT_CODE add_recursive_fields(php_http_message_body_t *body, cons
 	zval *val;
 	php_http_arrkey_t key;
 
-	if (!ZEND_HASH_GET_APPLY_COUNT(fields)) {
-		ZEND_HASH_INC_APPLY_COUNT(fields);
+	if (!HT_IS_RECURSIVE(fields)) {
+		HT_PROTECT_RECURSION(fields);
 		ZEND_HASH_FOREACH_KEY_VAL_IND(fields, key.h, key.key, val)
 		{
 			char *str = format_key(&key, name);
@@ -349,18 +358,18 @@ static ZEND_RESULT_CODE add_recursive_fields(php_http_message_body_t *body, cons
 			if (Z_TYPE_P(val) != IS_ARRAY && Z_TYPE_P(val) != IS_OBJECT) {
 				if (SUCCESS != add_recursive_field_value(body, str, val)) {
 					efree(str);
-					ZEND_HASH_DEC_APPLY_COUNT(fields);
+					HT_UNPROTECT_RECURSION(fields);
 					return FAILURE;
 				}
 			} else if (SUCCESS != add_recursive_fields(body, str, HASH_OF(val))) {
 				efree(str);
-				ZEND_HASH_DEC_APPLY_COUNT(fields);
+				HT_UNPROTECT_RECURSION(fields);
 				return FAILURE;
 			}
 			efree(str);
 		}
 		ZEND_HASH_FOREACH_END();
-		ZEND_HASH_DEC_APPLY_COUNT(fields);
+		HT_UNPROTECT_RECURSION(fields);
 	}
 
 	return SUCCESS;
@@ -378,23 +387,26 @@ static ZEND_RESULT_CODE add_recursive_files(php_http_message_body_t *body, const
 		zval *val;
 		php_http_arrkey_t key;
 
-		if (!ZEND_HASH_GET_APPLY_COUNT(files)) {
-			ZEND_HASH_INC_APPLY_COUNT(files);
+		if (!HT_IS_RECURSIVE(files)) {
+			HT_PROTECT_RECURSION(files);
 			ZEND_HASH_FOREACH_KEY_VAL_IND(files, key.h, key.key, val)
 			{
 				if (Z_TYPE_P(val) == IS_ARRAY || Z_TYPE_P(val) == IS_OBJECT) {
-					char *str = format_key(&key, name);
+					char *str = key.key ? format_key(&key, name) : NULL;
+					const char *prefix = str ? str : name;
 
-					if (SUCCESS != add_recursive_files(body, str, HASH_OF(val))) {
+					if (SUCCESS != add_recursive_files(body, prefix, HASH_OF(val))) {
 						efree(str);
-						ZEND_HASH_DEC_APPLY_COUNT(files);
+						HT_UNPROTECT_RECURSION(files);
 						return FAILURE;
 					}
-					efree(str);
+					if (str) {
+						efree(str);
+					}
 				}
 			}
 			ZEND_HASH_FOREACH_END();
-			ZEND_HASH_DEC_APPLY_COUNT(files);
+			HT_UNPROTECT_RECURSION(files);
 		}
 		return SUCCESS;
 	} else {
@@ -408,7 +420,7 @@ static ZEND_RESULT_CODE add_recursive_files(php_http_message_body_t *body, const
 			} else {
 				zend_string *tmp = zval_get_string(zdata);
 
-				stream = php_stream_memory_open(TEMP_STREAM_READONLY, tmp->val, tmp->len);
+				stream = php_http_mem_stream_open(TEMP_STREAM_READONLY, tmp);
 				zend_string_release(tmp);
 			}
 		} else {
@@ -420,7 +432,7 @@ static ZEND_RESULT_CODE add_recursive_files(php_http_message_body_t *body, const
 			return FAILURE;
 		} else {
 			zend_string *znc = zval_get_string(zname), *ztc = zval_get_string(ztype);
-			php_http_arrkey_t arrkey = {0, znc};
+			php_http_arrkey_t arrkey = {0, znc, 0, 0};
 			char *key = format_key(&arrkey, name);
 			ZEND_RESULT_CODE ret = php_http_message_body_add_form_file(body, key, ztc->val, zfc->val, stream);
 
@@ -567,6 +579,8 @@ php_http_message_body_object_t *php_http_message_body_object_new_ex(zend_class_e
 	zend_object_std_init(&o->zo, php_http_message_body_class_entry);
 	object_properties_init(&o->zo, ce);
 
+	o->gc = emalloc(sizeof(zval));
+
 	if (body) {
 		o->body = body;
 	}
@@ -576,10 +590,10 @@ php_http_message_body_object_t *php_http_message_body_object_new_ex(zend_class_e
 	return o;
 }
 
-zend_object *php_http_message_body_object_clone(zval *object)
+zend_object *php_http_message_body_object_clone(zend_object *object)
 {
-	php_http_message_body_object_t *new_obj = NULL;
-	php_http_message_body_object_t *old_obj = PHP_HTTP_OBJ(NULL, object);
+	php_http_message_body_object_t *new_obj;
+	php_http_message_body_object_t *old_obj = PHP_HTTP_OBJ(object, NULL);
 	php_http_message_body_t *body = php_http_message_body_copy(old_obj->body, NULL);
 
 	new_obj = php_http_message_body_object_new_ex(old_obj->zo.ce, body);
@@ -588,10 +602,40 @@ zend_object *php_http_message_body_object_clone(zval *object)
 	return &new_obj->zo;
 }
 
+static HashTable *php_http_message_body_object_get_gc(zend_object *object, zval **table, int *n)
+{
+	php_http_message_body_object_t *obj = PHP_HTTP_OBJ(object, NULL);
+	HashTable *props = object->handlers->get_properties(object);
+	uint32_t count = zend_hash_num_elements(props);
+
+	obj->gc = erealloc(obj->gc, (1 + count) * sizeof(zval));
+
+	if (php_http_message_body_stream(obj->body)) {
+		*n = 1;
+		php_stream_to_zval(php_http_message_body_stream(obj->body), obj->gc);
+	} else {
+		*n = 0;
+	}
+
+	if (count) {
+		zval *val;
+
+		ZEND_HASH_FOREACH_VAL(props, val)
+		{
+			ZVAL_COPY_VALUE(&obj->gc[(*n)++], val);
+		}
+		ZEND_HASH_FOREACH_END();
+	}
+	*table = obj->gc;
+
+	return NULL;
+}
+
 void php_http_message_body_object_free(zend_object *object)
 {
 	php_http_message_body_object_t *obj = PHP_HTTP_OBJ(object, NULL);
 
+	PTR_FREE(obj->gc);
 	php_http_message_body_free(&obj->body);
 	zend_object_std_dtor(object);
 }
@@ -600,6 +644,7 @@ void php_http_message_body_object_free(zend_object *object)
 	do { \
 		if (!obj->body) { \
 			obj->body = php_http_message_body_init(NULL, NULL); \
+			php_stream_to_zval(php_http_message_body_stream(obj->body), obj->gc); \
 		} \
 	} while(0)
 
@@ -621,12 +666,13 @@ PHP_METHOD(HttpMessageBody, __construct)
 			php_http_message_body_free(&obj->body);
 		}
 		obj->body = php_http_message_body_init(NULL, stream);
+		php_stream_to_zval(stream, obj->gc);
 	}
 }
 
-ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessageBody___toString, 0, 0, 0)
+ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessageBody_serialize, 0, 0, 0)
 ZEND_END_ARG_INFO();
-PHP_METHOD(HttpMessageBody, __toString)
+PHP_METHOD(HttpMessageBody, serialize)
 {
 	if (SUCCESS == zend_parse_parameters_none()) {
 		php_http_message_body_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
@@ -647,17 +693,60 @@ ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessageBody_unserialize, 0, 0, 1)
 ZEND_END_ARG_INFO();
 PHP_METHOD(HttpMessageBody, unserialize)
 {
-	char *us_str;
-	size_t us_len;
+	zend_string *us_str;
 
-	if (SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "s", &us_str, &us_len)) {
+	if (SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "S", &us_str)) {
 		php_http_message_body_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
-		php_stream *s = php_stream_memory_open(0, us_str, us_len);
+		php_stream *s = php_http_mem_stream_open(0, us_str);
 
 		obj->body = php_http_message_body_init(NULL, s);
+		php_stream_to_zval(s, obj->gc);
 	}
 }
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessageBody___unserialize, 0, 1, IS_VOID, 0)
+	ZEND_ARG_TYPE_INFO(0, data, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+PHP_METHOD(HttpMessageBody, __unserialize)
+{
+	HashTable *arr;
+
+	if (SUCCESS == zend_parse_parameters(ZEND_NUM_ARGS(), "h", &arr)) {
+		zval *zv = zend_hash_index_find(arr, 0);
+
+		if (zv) {
+			zend_string *zs = zval_get_string(zv);
+			php_stream *s = php_http_mem_stream_open(0, zs);
+			php_http_message_body_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
+
+			obj->body = php_http_message_body_init(NULL, s);
+			php_stream_to_zval(s, obj->gc);
+			zend_string_release(zs);
+		}
+	}
+}
+
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessageBody___serialize, 0, 0, IS_ARRAY, 0)
+ZEND_END_ARG_INFO();
+PHP_METHOD(HttpMessageBody, __serialize)
+{
+
+	php_http_message_body_object_t *obj = PHP_HTTP_OBJ(NULL, getThis());
+	zend_string *zs;
+
+	zend_parse_parameters_none();
+
+	PHP_HTTP_MESSAGE_BODY_OBJECT_INIT(obj);
+
+	array_init(return_value);
+	zs = php_http_message_body_to_string(obj->body, 0, 0);
+	if (zs) {
+		add_index_str(return_value, 0, zs);
+	}
+}
+
+
+
 ZEND_BEGIN_ARG_INFO_EX(ai_HttpMessageBody_toStream, 0, 0, 1)
 	ZEND_ARG_INFO(0, stream)
 	ZEND_ARG_INFO(0, offset)
@@ -712,8 +801,8 @@ PHP_METHOD(HttpMessageBody, getResource)
 
 		PHP_HTTP_MESSAGE_BODY_OBJECT_INIT(obj);
 
-		++GC_REFCOUNT(obj->body->res);
-		RETVAL_RES(obj->body->res);
+		php_stream_to_zval(php_http_message_body_stream(obj->body), return_value);
+		Z_ADDREF_P(return_value);
 	}
 }
 
@@ -862,12 +951,17 @@ PHP_METHOD(HttpMessageBody, stat)
 	}
 }
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_HttpMessageBody___toString, 0, 0, IS_STRING, 0)
+ZEND_END_ARG_INFO();
+
 static zend_function_entry php_http_message_body_methods[] = {
-	PHP_ME(HttpMessageBody, __construct,  ai_HttpMessageBody___construct,  ZEND_ACC_PUBLIC|ZEND_ACC_CTOR)
-	PHP_ME(HttpMessageBody, __toString,   ai_HttpMessageBody___toString,   ZEND_ACC_PUBLIC)
-	PHP_MALIAS(HttpMessageBody, toString, __toString, ai_HttpMessageBody___toString, ZEND_ACC_PUBLIC)
-	PHP_MALIAS(HttpMessageBody, serialize, __toString, ai_HttpMessageBody___toString, ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessageBody, __construct,  ai_HttpMessageBody___construct,  ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessageBody, serialize,    ai_HttpMessageBody_serialize,    ZEND_ACC_PUBLIC)
+	PHP_MALIAS(HttpMessageBody, toString, serialize, ai_HttpMessageBody_serialize, ZEND_ACC_PUBLIC)
+	PHP_MALIAS(HttpMessageBody, __toString, serialize, ai_HttpMessageBody___toString, ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessageBody, unserialize,  ai_HttpMessageBody_unserialize,  ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessageBody, __serialize,  ai_HttpMessageBody___serialize,  ZEND_ACC_PUBLIC)
+	PHP_ME(HttpMessageBody, __unserialize,ai_HttpMessageBody___unserialize,ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessageBody, toStream,     ai_HttpMessageBody_toStream,     ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessageBody, toCallback,   ai_HttpMessageBody_toCallback,   ZEND_ACC_PUBLIC)
 	PHP_ME(HttpMessageBody, getResource,  ai_HttpMessageBody_getResource,  ZEND_ACC_PUBLIC)
@@ -891,6 +985,7 @@ PHP_MINIT_FUNCTION(http_message_body)
 	php_http_message_body_object_handlers.offset = XtOffsetOf(php_http_message_body_object_t, zo);
 	php_http_message_body_object_handlers.clone_obj = php_http_message_body_object_clone;
 	php_http_message_body_object_handlers.free_obj = php_http_message_body_object_free;
+	php_http_message_body_object_handlers.get_gc = php_http_message_body_object_get_gc;
 	zend_class_implements(php_http_message_body_class_entry, 1, zend_ce_serializable);
 
 	return SUCCESS;
